@@ -2,7 +2,7 @@ import time
 import numpy as np
 import numba
 import warnings
-import kernelml
+from kernelml import KernelML
 import sys
 import copy
 from functools import partial
@@ -52,13 +52,12 @@ class DensityFactorization():
                 data,_ = hdre_histogram(_X_,y,agg_func='count',cost=None,bins=bins3)
                 data=data/np.sum(data)
                 new_kde = np.fft.fftshift(np.real(np.fft.ifftn(np.fft.fftn(data)*self.fftkernel)))
-                self.kde_target[count]=self.kde_target[count]*alpha + new_kde*(1-alpha)
+                self.kde_target[count]=self.kde_target[count]*(1-alpha) + new_kde*(alpha)
                 count+=1
 
-    def optimize(self,X,y=None,cost=None, number_of_random_simulations=1000, number_of_realizations=10, agg_func='count',verbose=True):
-                 
-
-            
+    def optimize(self,X,y=None,cost=None, number_of_random_simulations=1000, number_of_realizations=10, agg_func='count',verbose=True,
+        dim_tuples=None
+        ):
         
         self.simulations = number_of_random_simulations
         self.realizations = number_of_realizations
@@ -78,7 +77,12 @@ class DensityFactorization():
         widths = (maxs-mins)/(self.bins_per_dim)
         
         self.dim_bins = [np.linspace(m0,m1,self.bins_per_dim) for m0,m1 in zip(mins,maxs)]
-        self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
+        
+        #custom dimensional pair TODO
+        if dim_tuples is None:
+            self.dim_combos = [(i,j) for i in range(X.shape[1]) for j in range(X.shape[1]) if j>i]
+        else:
+            self.dim_combos=dim_tuples
         
         combo_len = len(self.dim_combos)
         
@@ -94,6 +98,7 @@ class DensityFactorization():
                 data,_ = hdre_histogram(_X_,y,agg_func=agg_func,cost=cost,bins=bins3)
                 data=data/np.sum(data)
                 self.kde_target[count] = np.fft.fftshift(np.real(np.fft.ifftn(np.fft.fftn(data)*self.fftkernel)))
+
                 count+=1
 
         
@@ -115,7 +120,7 @@ class DensityFactorization():
         self.args = args
         
         if self.kmldata is None:
-            self.kml = kernelml.KernelML(
+            self.kml = KernelML(
                  prior_sampler_fcn=hdre_prior_sampler,
                  posterior_sampler_fcn=None,
                  intermediate_sampler_fcn=None,
@@ -128,7 +133,6 @@ class DensityFactorization():
             self.kml.load_kmldata(self.kmldata)
 #            self.kml.kmldata.args = self.args
             
-
         self.kml.optimize(X[:1],np.array([[]]),
                                         convergence_z_score=3.0,
                                         min_loss_per_change=0.0,
@@ -279,27 +283,39 @@ class DensityFactorization():
         return loss_matrix
 
 
-class HeirarchicalDensityFactorization():
+class HierarchicalDensityFactorization():
     
-    def __init__(self,num_clusters=6,
+    def __init__(self,num_clusters=8,
                  bins_per_dimension=61,
-                 smoothing_parameter=3.0,
+                 smoothing_parameter=1.0,
+                 min_leaf_samples=1,
+                 alpha = 0.1,
                  verbose=0):
+        
+        self.alpha = alpha
+        self.min_leaf_samples = min_leaf_samples
         self.assignments = []
         self.centroids = []
+        self.deviations = []
+        self.min_losses = []
         self.models = []
         self.burn_in = 10
         self.verbose = verbose
+        self.kde_true = []
+        self.kde_pred = []
+        self.KDE = None
         
         self.smoothing_parameter = smoothing_parameter
         
-        self.model = kernelml.region_estimator.DensityFactorization(number_of_clusters=num_clusters,
+        self.model = DensityFactorization(number_of_clusters=num_clusters,
                                                                     bins_per_dimension=bins_per_dimension,
                                                                     smoothing_parameter=smoothing_parameter)
         self.count = 0
         
         
-    def assign(self,X,pad=1.):
+ 
+    def refactor_assignments(self,X,pad=1.):
+        
         assignments = []
         for model in self.models:
             
@@ -307,36 +323,75 @@ class HeirarchicalDensityFactorization():
             
             Z = np.diag(np.dot(D.T,D))
             for i in range(Z.shape[0]):
-                if Z[i]>0:
+                if Z[i]>self.min_leaf_samples:
                     assignments.append(D[:,i:i+1])
+                    
+#             E = np.hstack(self.assignments)
+#             M = (np.sum(E,axis=1)==0)
 
-        return np.hstack(assignments)
+#             alpha = np.mean(M)*self.alpha
+
+#             self.model.update_kde(X[M],X[M],alpha)
+            
+        return assignments
+
+#    def assign(self,X,pad=1.):
+#        assignments = []
+#        for model in self.models:
+#
+#            D = model.get_assignments(X,pad=1.).astype(np.int)
+#
+#            Z = np.diag(np.dot(D.T,D))
+#            for i in range(Z.shape[0]):
+#                if Z[i]>self.min_leaf_samples:
+#                    assignments.append(D[:,i:i+1])
+#
+#        return np.hstack(assignments)
+       
+    def assign(self,X,pad=1):
+        num_clusters = np.vstack(self.centroids).shape[0]
+        mask = np.zeros((X.shape[0],num_clusters),dtype=np.bool)
+        k=0
+        for C,S in zip(self.centroids,self.deviations):
+            for i in range(C.shape[0]):
+                lower = C[i].flatten()-S.flatten()*pad
+                upper = C[i].flatten()+S.flatten()*pad
+                mask[:,k] = np.all((X>=lower)&(X<=upper),axis=1)
+                k+=1
+                
+        return mask
     
     def factorize(self,X,M):
 
-        self.count+=1
         
         """
         alpha controls how much weight to give to data points not assign to a cluster
         """
         
-        alpha = 1-np.mean(M)
+        alpha = np.mean(M)*self.alpha
 
         self.model.update_kde(X[M],X[M],alpha)
         
-        realizations = 5
-        if self.count==1:
-            realizations = self.burn_in
+        realizations = self.realizations
             
         self.model.kmldata=None
         
-        self.model.optimize(X=X,y=X,  number_of_random_simulations=125,
-                   number_of_realizations=realizations,
-                     agg_func='count',
-                     verbose=self.verbose)
+#         model.prune_clusters(y,pad=2.0,limit=1)
+        
+        self.model.optimize(X=X,y=X,  number_of_random_simulations=self.number_of_random_simulations,
+                   number_of_realizations = realizations,
+                   agg_func='count',
+                   verbose=self.verbose,
+                   dim_tuples=self.dim_tuples
+                     )
+        
+#         if self.count==0:
+#             self.KDE = self.model.kde_target.copy()
+        
+        self.count+=1
         
-        """
         #this will plot the estimated kernel density estimated
+        """
         count=0
         for i,j in self.model.dim_combos:
 
@@ -350,45 +405,129 @@ class HeirarchicalDensityFactorization():
             plt.show()
             count+=1
         """
+            
+        self.kde_true.append(self.model.kde_target.copy())
+        self.kde_pred.append(self.model.kde_estimate.copy())
+        
 
         D = self.model.get_assignments(X,pad=1.).astype(np.int)
         C = self.model.centroids_
+        S = self.model.deviations_
 
+        minloss = self.model.kmldata.loss_by_realization[-1][0]
+        self.min_losses.append(minloss)
+        
         self.models.append(copy.copy(self.model))
         
         #add data points to the collected list
         Z = np.diag(np.dot(D.T,D))
+        
+        if self.verbose==True:
+            print(np.sum(Z==0),'clusters were not assigned data points\n')
+        
+        assigns = []
+        centers = []
         for i in range(Z.shape[0]):
             #only add if the cluster has new data points?
-            if Z[i]>0:
-                self.assignments.append(D[:,i:i+1])
-                self.centroids.append(C[:,i:i+1])
-
+            if Z[i]>self.min_leaf_samples:
+                assigns.append(D[:,i:i+1])
+                centers.append(C[i:i+1])
+                
+                
+        self.assignments.append(assigns)
+        self.centroids.append(np.vstack(centers))
+        self.deviations.append(S)
             
-    def optimize(self,X,maxiter=20,verbose=0):
+    def optimize(self,X,
+                 maxiter=20,
+                 stop_thrshld=0.01,
+                 realizations=10,
+                 number_of_random_simulations=200,
+                 dim_tuples=None,
+                 verbose=0):
         
-        self.verbose=verbose
+        self.verbose = verbose
+        self.dim_tuples = dim_tuples
+        self.realizations = realizations
+        self.number_of_random_simulations = number_of_random_simulations
         M = np.ones(X.shape[0],dtype=bool)
         LAST = M
         for _ in range(maxiter):
+            self.factorize(X,M)
+            E = np.hstack(sum(self.assignments,[]))
+            M = (np.sum(E,axis=1)==0)
             if self.verbose==True:
                 print(np.sum(M),'data points are unassigned')
-            self.factorize(X,M)
-            E = np.hstack(self.assignments)
-            M = (np.sum(E,axis=1)==0)
             
             #break when no data points are left
             if np.sum(M)==0:
                 break
                 
-            #break if no new data points are assigned
-            if np.sum(LAST)-np.sum(M)==0:
+            #break if no threshold is met
+                        
+            
+            prune_losses = []
+            prune_kde_true = []
+            prune_kde_pred = []
+            prune_models = []
+            prune_centroids = []
+            prune_deviations = []
+            prune_assignments = []
+            keep_count=0
+            
+            if len(self.min_losses)<=1:
+                continue
+            
+            
+            for i,loss in enumerate(self.min_losses):
+                if loss*0.5<=np.min(self.min_losses):
+                    prune_models.append(self.models[i])
+                    prune_kde_true.append(self.kde_true[i])
+                    prune_kde_pred.append(self.kde_pred[i])
+                    prune_losses.append(self.min_losses[i])
+                    prune_centroids.append(self.centroids[i])
+                    prune_deviations.append(self.deviations[i])
+                    prune_assignments.append(self.assignments[i])
+                    keep_count+=1
+                    
+            prune_count = len(self.min_losses)-keep_count
+
+            if self.verbose==True and prune_count>0:
+                print('pruning {} iterations\n'.format(prune_count))
+                
+            self.models = prune_models
+            self.kde_true = prune_kde_true
+            self.kde_pred = prune_kde_pred
+            self.min_losses = prune_losses
+            self.centroids = prune_centroids
+            self.deviations = prune_deviations
+            self.assignments = prune_assignments
+            
+
+#            if prune_count>0:
+#                self.assignments = self.refactor_assignments(X)
+            
+            
+            
+            if np.mean(M)<stop_thrshld and prune_count==0:
+                if self.verbose==True:
+                    print('stopping with {} unassigned data points'.format(np.sum(M)))
+                break
+            
+            if (np.sum(LAST)-np.sum(M))==0 and prune_count==0:
                 if self.verbose==True:
                     print('stopping with {} unassigned data points'.format(np.sum(M)))
                 break
                 
             LAST = M
             
+        if self.verbose==True and self.count==maxiter:
+            print('stopping at {} iterations'.format(maxiter))
+        
+
+            
+        
+            
 
 @jit('float64[:,:](float64[:,:], float64,float64)',nopython=True)
 def uniform_kernel(x,mu_,var_):
@@ -403,7 +542,7 @@ def uniform_kernel(x,mu_,var_):
     if var<width:
         var = width
     
-    pdf = ((x+width/2>=mu-var)&(x+width/2<=mu+var)).reshape(-1,1)
+    pdf = ((x+width>=mu-var)&(x+width<=mu+var)).reshape(-1,1)
     pdf = pdf.astype(np.float64)
     
     if x[1]>mu-var:
@@ -443,14 +582,17 @@ def hdre_loss(X,y,w_,dim_combos, pdf_combos,bin_combos, min_lim, max_lim, param_
              bins1 = bins[:,:,1].astype(np.float64)
  #             uniform_kernel(bins0,w[k,i],var[i,0])
              pdf = uniform_kernel(bins0,w[k,i],var[i,0]).dot(uniform_kernel(bins1,w[k,j],var[j,0]).T)
+          
  #             print(pdf.shape)
              data1 += pdf
+             
 
          pdf1=data1/np.sum(data1)
 
          pdf2 = pdf_combos[count]
 
          diff = (pdf2-pdf1).flatten()*100
+         
          ll += np.sum(diff**2)
 
          count+=1
